@@ -1,12 +1,12 @@
 /*
-**  Copyright (c) 2004, 2005 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2004-2006 Sendmail, Inc. and its suppliers.
 **	All rights reserved.
 **
-**  $Id: dk-filter.c,v 1.137 2005/04/28 23:47:56 msk Exp $
+**  $Id: dk-filter.c,v 1.158 2006/05/19 21:42:05 msk Exp $
 */
 
 #ifndef lint
-static char dk_filter_c_id[] = "@(#)$Id: dk-filter.c,v 1.137 2005/04/28 23:47:56 msk Exp $";
+static char dk_filter_c_id[] = "@(#)$Id: dk-filter.c,v 1.158 2006/05/19 21:42:05 msk Exp $";
 #endif /* !lint */
 
 /* system includes */
@@ -40,6 +40,8 @@ static char dk_filter_c_id[] = "@(#)$Id: dk-filter.c,v 1.137 2005/04/28 23:47:56
 #include <pthread.h>
 #include <netdb.h>
 #include <signal.h>
+#include <regex.h>
+#include <openssl/err.h>
 #if _FFR_REPORTINFO
 # include <openssl/bio.h>
 # include <openssl/evp.h>
@@ -52,9 +54,6 @@ static char dk_filter_c_id[] = "@(#)$Id: dk-filter.c,v 1.137 2005/04/28 23:47:56
 #else /* SOLARIS */
 # include <paths.h>
 #endif /* SOLARIS */
-#if _FFR_MULTIPLE_KEYS
-# include <regex.h>
-#endif /* _FFR_MULTIPLE_KEYS */
 
 /* sendmail includes */
 #include <sm/cdefs.h>
@@ -284,7 +283,6 @@ sfsistat mlfi_eoh __P((SMFICTX *));
 sfsistat mlfi_eom __P((SMFICTX *));
 sfsistat mlfi_header __P((SMFICTX *, char *, char *));
 
-static char *dkf_stringlookup __P((int val, struct lookup *table));
 static void dkf_cleanup __P((SMFICTX *));
 
 /* GLOBALS */
@@ -319,15 +317,13 @@ struct keytable *keytail;			/* key list */
 #endif /* _FFR_MULTIPLE_KEYS */
 Peer peerlist;					/* queue of "peers" */
 Peer internal;					/* queue of "internal" hosts */
-#if _FFR_EXTERNAL_IGNORE_LIST
 Peer exignore;					/* "external ignore" hosts */
-#endif /* _FFR_EXTERNAL_IGNORE_LIST */
 char **domains;					/* domains to sign */
+regex_t **dompats;				/* domain patterns */
 char **mtas;					/* MTA ports to sign */
-#if _FFR_MACRO_LIST
+char **omithdrs;				/* headers to omit */
 char **macros;					/* macros/values to check */
 char **values;					/* macros/values to check */
-#endif /* _FFR_MACRO_LIST */
 #if POPAUTH
 DB *popdb;					/* POP auth DB */
 #endif /* POPAUTH */
@@ -421,7 +417,7 @@ dkf_killchild(pid_t pid, int sig)
 */
 
 static void
-dkf_zapkey()
+dkf_zapkey(void)
 {
 	if (seckey != NULL)
 	{
@@ -454,7 +450,7 @@ dkf_zapkey()
 */
 
 static void
-dkf_stdio()
+dkf_stdio(void)
 {
 	int devnull;
 
@@ -567,7 +563,7 @@ dkf_queryalg(int alg)
 */
 
 static msgctx
-dkf_initcontext()
+dkf_initcontext(void)
 {
 	msgctx ctx;
 
@@ -596,8 +592,7 @@ dkf_initcontext()
 */
 
 static void
-dkf_cleanup(ctx)
-	SMFICTX *ctx;
+dkf_cleanup(SMFICTX *ctx)
 {
 	msgctx dfc;
 	connctx cc;
@@ -666,28 +661,6 @@ dkf_configlookup(char *opt, struct lookup *table)
 	}
 }
 
-/*
-**  DKF_STRINGLOOKUP -- look up the string for a config code or value
-**
-**  Parameters:
-**  	val -- value to look up
-**  	table -- lookup table to use
-**
-**  Return value:
-**  	Pointer to the matching string, or NULL on error.
-*/
-
-static char *
-dkf_stringlookup(int val, struct lookup *table)
-{
-	int c;
-
-	for (c = 0; ; c++)
-	{
-		if (table[c].code == -1 || table[c].code == val)
-		    return table[c].str;
-	}
-}
 /*
 **  DKF_PARSECONFIG -- parse configuration and/or apply defaults
 **
@@ -704,6 +677,7 @@ dkf_parseconfig(char *confstr)
 	int vs;
 	char *p;
 	char *v;
+	char *tmp;
 
 	/* load defaults */
 	memcpy(&config, &defaults, sizeof(config));
@@ -711,8 +685,16 @@ dkf_parseconfig(char *confstr)
 	if (confstr == NULL)
 		return TRUE;
 
+	tmp = strdup(confstr);
+	if (tmp == NULL)
+	{
+		fprintf(stderr, "%s: strdup(): %s\n", progname,
+		        strerror(errno));
+		return FALSE;
+	}
+
 	/* process configuration */
-	for (p = strtok(confstr, ","); p != NULL; p = strtok(NULL, ","))
+	for (p = strtok(tmp, ","); p != NULL; p = strtok(NULL, ","))
 	{
 		v = strchr(p, '=');
 		if (v == NULL)
@@ -808,7 +790,6 @@ dkf_libstatus(SMFICTX *ctx, char *where, int status)
 			       where == NULL ? "" : where,
 			       where == NULL ? "" : ": ", strerror(errno));
 		}
-		dkf_cleanup(ctx);
 		break;
 
 	  case DK_STAT_BADSIG:
@@ -818,7 +799,6 @@ dkf_libstatus(SMFICTX *ctx, char *where, int status)
 			syslog(LOG_ERR, "%s: bad signature data",
 			       JOBID(dfc->mctx_jobid));
 		}
-		dkf_cleanup(ctx);
 		break;
 
 	  case DK_STAT_NOSIG:
@@ -828,7 +808,6 @@ dkf_libstatus(SMFICTX *ctx, char *where, int status)
 			syslog(LOG_ERR, "%s: no signature data",
 			       JOBID(dfc->mctx_jobid));
 		}
-		dkf_cleanup(ctx);
 		break;
 
 	  case DK_STAT_NORESOURCE:
@@ -840,7 +819,6 @@ dkf_libstatus(SMFICTX *ctx, char *where, int status)
 			       where == NULL ? "" : where,
 			       where == NULL ? "" : ": ", strerror(errno));
 		}
-		dkf_cleanup(ctx);
 		break;
 
 	  case DK_STAT_CANTVRFY:
@@ -850,7 +828,6 @@ dkf_libstatus(SMFICTX *ctx, char *where, int status)
 			syslog(LOG_ERR, "%s: signature verification failed",
 			       JOBID(dfc->mctx_jobid));
 		}
-		dkf_cleanup(ctx);
 		break;
 
 	  case DK_STAT_SYNTAX:
@@ -860,10 +837,38 @@ dkf_libstatus(SMFICTX *ctx, char *where, int status)
 			syslog(LOG_ERR, "%s: syntax error in signature data",
 			       JOBID(dfc->mctx_jobid));
 		}
-		dkf_cleanup(ctx);
 		break;
 	}
 
+#ifdef _FFR_LOG_SSL_ERRORS
+	if (ERR_peek_error() != 0 && dolog)
+	{
+		int n;
+		unsigned long e;
+		char errbuf[BUFRSZ + 1];
+		char tmp[BUFRSZ + 1];
+
+		memset(errbuf, '\0', sizeof errbuf);
+		for (n = 0; ; n++)
+		{
+			e = ERR_get_error();
+			if (e == 0)
+				break;
+
+			memset(tmp, '\0', sizeof tmp);
+			(void) ERR_error_string_n(e, tmp, sizeof tmp);
+			if (n != 0)
+				sm_strlcat(errbuf, "; ", sizeof errbuf);
+			sm_strlcat(errbuf, tmp, sizeof errbuf);
+		}
+
+		syslog(LOG_INFO, "%s SSL %s", JOBID(dfc->mctx_jobid), errbuf);
+	}
+#endif /* _FFR_LOG_SSL_ERRORS */
+
+	if (status != DK_STAT_OK)
+		dkf_cleanup(ctx);
+
 	return retcode;
 }
 
@@ -1321,10 +1326,7 @@ dkf_report(msgctx dfc, char *result, char *reason)
 */
 
 sfsistat
-mlfi_connect(ctx, host, ip)
-	SMFICTX *ctx;
-	char *host;
-	_SOCK_ADDR *ip;
+mlfi_connect(SMFICTX *ctx, char *host, _SOCK_ADDR *ip)
 {
 	connctx cc;
 
@@ -1405,9 +1407,7 @@ mlfi_connect(ctx, host, ip)
 */
 
 sfsistat
-mlfi_envfrom(ctx, envfrom)
-	SMFICTX *ctx;
-	char **envfrom;
+mlfi_envfrom(SMFICTX *ctx, char **envfrom)
 {
 	connctx cc;
 	msgctx dfc;
@@ -1469,10 +1469,7 @@ mlfi_envfrom(ctx, envfrom)
 */
 
 sfsistat
-mlfi_header(ctx, headerf, headerv)
-	SMFICTX *ctx;
-	char *headerf;
-	char *headerv;
+mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
 {
 	msgctx dfc;
 	connctx cc;
@@ -1552,12 +1549,12 @@ mlfi_header(ctx, headerf, headerv)
 */
 
 sfsistat
-mlfi_eoh(ctx)
-	SMFICTX *ctx;
+mlfi_eoh(SMFICTX *ctx)
 {
 	bool msgsigned;
 	bool domainok;
 	bool originok;
+	bool skip;
 	int status;
 	connctx cc;
 	msgctx dfc;
@@ -1606,12 +1603,6 @@ mlfi_eoh(ctx)
 		    dkf_findheader(dfc, "From", 1) != NULL)
 			ok = FALSE;
 
-		/* either 0 or 1 To: or Cc: */
-		if (dkf_findheader(dfc, "To", 1) != NULL)
-			ok = FALSE;
-		if (dkf_findheader(dfc, "Cc", 1) != NULL)
-			ok = FALSE;
-
 		/* exactly one Date: */
 		if (dkf_findheader(dfc, "Date", 0) == NULL ||
 		    dkf_findheader(dfc, "Date", 1) != NULL)
@@ -1683,13 +1674,31 @@ mlfi_eoh(ctx)
 	{
 		int n;
 
-		for (n = 0; domains[n] != NULL; n++)
+		if (dompats != NULL)
 		{
-			if (strcasecmp(dfc->mctx_domain, domains[n]) == 0)
+			for (n = 0; dompats[n] != NULL; n++)
 			{
-				dfc->mctx_domain = domains[n];
-				domainok = TRUE;
-				break;
+				status = regexec(dompats[n],
+				                 dfc->mctx_domain,
+				                 0, NULL, 0);
+				if (status == 0)
+				{
+					domainok = TRUE;
+					break;
+				}
+			}
+		}
+		else
+		{
+			for (n = 0; domains[n] != NULL; n++)
+			{
+				if (strcasecmp(dfc->mctx_domain,
+				               domains[n]) == 0)
+				{
+					dfc->mctx_domain = domains[n];
+					domainok = TRUE;
+					break;
+				}
 			}
 		}
 
@@ -1704,13 +1713,31 @@ mlfi_eoh(ctx)
 				if (*p == '\0')
 					break;
 
-				for (n = 0; domains[n] != NULL; n++)
+				if (dompats != NULL)
 				{
-					if (strcasecmp(p,  domains[n]) == 0)
+					for (n = 0; dompats[n] != NULL; n++)
 					{
-						dfc->mctx_domain = domains[n];
-						domainok = TRUE;
-						break;
+						status = regexec(dompats[n],
+						                 dfc->mctx_domain,
+						                 0, NULL, 0);
+						if (status == 0)
+						{
+							domainok = TRUE;
+							break;
+						}
+					}
+				}
+				else
+				{
+					for (n = 0; domains[n] != NULL; n++)
+					{
+						if (strcasecmp(p,
+						               domains[n]) == 0)
+						{
+							dfc->mctx_domain = domains[n];
+							domainok = TRUE;
+							break;
+						}
 					}
 				}
 			}
@@ -1784,7 +1811,6 @@ mlfi_eoh(ctx)
 		}
 	}
 
-#if _FFR_MACRO_LIST
 	/* see if macro tests passed */
 	if (macros != NULL)
 	{
@@ -1807,7 +1833,6 @@ mlfi_eoh(ctx)
 			}
 		}
 	}
-#endif /* _FFR_MACRO_LIST */
 
 	/* see if it came from an internal or authenticated source */
 	if (!originok)
@@ -1822,13 +1847,9 @@ mlfi_eoh(ctx)
 		    !dkf_checkhost(internal, cc->cctx_host) &&
 		    !dkf_checkip(internal, &cc->cctx_ip))
 		{
-#if _FFR_EXTERNAL_IGNORE_LIST
 			if (domainok && dolog &&
 			    !dkf_checkhost(exignore, cc->cctx_host) &&
 			    !dkf_checkip(exignore, &cc->cctx_ip))
-#else /* _FFR_EXTERNAL_IGNORE_LIST */
-			if (domainok && dolog)
-#endif /* _FFR_EXTERNAL_IGNORE_LIST */
 			{
 				syslog(LOG_NOTICE,
 				       "%s external host %s attempted to send as %s",
@@ -1898,6 +1919,25 @@ mlfi_eoh(ctx)
 	memset(hdrbuf, '\0', sizeof hdrbuf);
 	for (hdr = dfc->mctx_hqhead; hdr != NULL; hdr = hdr->hdr_next)
 	{
+		skip = FALSE;
+
+		if (omithdrs != NULL && dfc->mctx_signing)
+		{
+			int c;
+
+			for (c = 0; omithdrs[c] != NULL; c++)
+			{
+				if (strcasecmp(hdr->hdr_hdr, omithdrs[c]) == 0)
+				{
+					skip = TRUE;
+					break;
+				}
+			}
+
+			if (skip)
+				continue;
+		}
+
 		snprintf(hdrbuf, MAXHEADER, "%s: %s%s", hdr->hdr_hdr,
 		         hdr->hdr_val, CRLF);
 		status = dk_header(dfc->mctx_dk, hdrbuf, strlen(hdrbuf));
@@ -1960,10 +2000,7 @@ mlfi_eoh(ctx)
 */
 
 sfsistat
-mlfi_body(ctx, bodyp, bodylen)
-	SMFICTX *ctx;
-	u_char *bodyp;
-	size_t bodylen;
+mlfi_body(SMFICTX *ctx, u_char *bodyp, size_t bodylen)
 {
 	int status;
 	msgctx dfc;
@@ -2009,8 +2046,7 @@ mlfi_body(ctx, bodyp, bodylen)
 */
 
 sfsistat
-mlfi_eom(ctx)
-	SMFICTX *ctx;
+mlfi_eom(SMFICTX *ctx)
 {
 	int status = DK_STAT_OK;
 	int c;
@@ -2256,20 +2292,22 @@ mlfi_eom(ctx)
 			                      val, sizeof val);
 
 			snprintf(header, sizeof header,
-			         "%s %s=%s; domainkeys=%s%s%s%s",
+			         "%s %s=%s; domainkeys=%s%s%s%s%s",
 			         hostname, hdr, val, authresult,
 			         comment == NULL ? "" : " (",
 			         comment == NULL ? "" : comment,
-			         comment == NULL ? "" : ")");
+			         comment == NULL ? "" : ")",
+                                 !(dkf & DK_FLAG_TESTING) ? "" : " (testing)");
 		}
 		else
 		{
 			snprintf(header, sizeof header,
-			         "%s; domainkeys=%s%s%s%s",
+			         "%s; domainkeys=%s%s%s%s%s",
 			         hostname, authresult,
 			         comment == NULL ? "" : " (",
 			         comment == NULL ? "" : comment,
-			         comment == NULL ? "" : ")");
+			         comment == NULL ? "" : ")",
+                                 !(dkf & DK_FLAG_TESTING) ? "" : " (testing)");
 		}
 
 		if (smfi_insheader(ctx, 1, AUTHRESULTSHDR,
@@ -2284,7 +2322,8 @@ mlfi_eom(ctx)
 		}
 
 # if _FFR_REPORTINFO
-		if (dfc->mctx_status != DKF_STATUS_GOOD)
+		if (dfc->mctx_status == DKF_STATUS_BAD &&
+		    dfc->mctx_dk != NULL)
 			dkf_report(dfc, authresult, comment);
 # endif /* _FFR_REPORTINFO */
 	}
@@ -2356,6 +2395,9 @@ mlfi_eom(ctx)
 		break;
 	}
 
+        if ((dkf & DK_FLAG_TESTING) != 0)
+                ret = SMFIS_ACCEPT;
+
 	return ret;
 }
 
@@ -2371,8 +2413,7 @@ mlfi_eom(ctx)
 */
 
 sfsistat
-mlfi_abort(ctx)
-	SMFICTX *ctx;
+mlfi_abort(SMFICTX *ctx)
 {
 	if (DK_DEBUG('t'))
 		syslog(LOG_INFO, "thread %p abort", pthread_self());
@@ -2392,11 +2433,14 @@ mlfi_abort(ctx)
 */
 
 sfsistat
-mlfi_close(ctx)
-	SMFICTX *ctx;
+mlfi_close(SMFICTX *ctx)
 {
 	connctx cc;
 
+#ifndef DEBUG
+	assert(ctx != NULL);
+#endif /* !DEBUG */
+
 	if (DK_DEBUG('t'))
 	{
 		pthread_mutex_lock(&count_lock);
@@ -2440,11 +2484,7 @@ struct smfiDesc smfilter =
 
 #ifdef DEBUG
 int
-smfi_chgheader(ctx, hdr, idx, val)
-	void *ctx;
-	char *hdr;
-	int idx;
-	char *val;
+smfi_chgheader(void *ctx, char *hdr, int idx, char *val)
 {
 	printf("smfi_chgheader(<ctx>, `%s', `%d', `%s')\n", hdr, idx,
 	       val == NULL ? "(null)" : val);
@@ -2452,10 +2492,7 @@ smfi_chgheader(ctx, hdr, idx, val)
 }
 
 int
-smfi_replacebody(ctx, p, len)
-	void *ctx;
-	char *p;
-	size_t len;
+smfi_replacebody(void *ctx, char *p, size_t len)
 {
 	printf("smfi_replacebody(<ctx>, `%.20s%s', `%d')\n", p,
 	       strlen(p) > 20 ? "..." : "", len);
@@ -2463,63 +2500,46 @@ smfi_replacebody(ctx, p, len)
 }
 
 int
-smfi_addheader(ctx, hdr, val)
-	void *ctx;
-	char *hdr;
-	char *val;
+smfi_addheader(void *ctx, char *hdr, char *val)
 {
 	printf("smfi_addheader(<ctx>, `%s', `%s')\n", hdr, val);
 	return MI_SUCCESS;
 }
 
 int
-smfi_insheader(ctx, idx, hdr, val)
-	void *ctx;
-	int idx;
-	char *hdr;
-	char *val;
+smfi_insheader(void *ctx, int idx, char *hdr, char *val)
 {
 	printf("smfi_insheader(<ctx>, %d, `%s', `%s')\n", idx, hdr, val);
 	return MI_SUCCESS;
 }
 
 void
-smfi_setconn(file)
-	char *file;
+smfi_setconn(char *file)
 {
 	printf("smfi_setconn(`%s')\n", file);
 }
 
 void
-smfi_setpriv(ctx, priv)
-	void *ctx;
-	void *priv;
+smfi_setpriv(void *ctx, void *priv)
 {
 	fakepriv = priv;
 }
 
 void *
-smfi_getpriv(ctx)
-	void *ctx;
+smfi_getpriv(void *ctx)
 {
 	return fakepriv;
 }
 
 void
-smfi_setreply(ctx, sc, esc, reply)
-	void *ctx;
-	char *sc;
-	char *esc;
-	char *reply;
+smfi_setreply(void *ctx, char *sc, char *esc, char *reply)
 {
 	printf("smfi_setreply(<ctx>, `%s', `%s', `%s')\n",
 	       sc, esc, reply);
 }
 
 char *
-smfi_getsymval(ctx, sym)
-	void *ctx;
-	char *sym;
+smfi_getsymval(void *ctx, char *sym)
 {
 	char *ret;
 	size_t l;
@@ -2545,7 +2565,7 @@ smfi_getsymval(ctx, sym)
 */
 
 int
-dkf_debug()
+dkf_debug(void)
 {
 	bool done;
 	int status;
@@ -2677,6 +2697,10 @@ dkf_debug()
 
 	status = mlfi_eom(NULL);
 	printf("mlfi_eom(NULL) returns %s\n", smfis_ret[status]);
+
+	status = mlfi_close(NULL);
+	printf("mlfi_close(NULL) returns %s\n", smfis_ret[status]);
+
 	return 0;
 }
 #endif /* DEBUG */
@@ -2692,7 +2716,7 @@ dkf_debug()
 */
 
 static int
-usage()
+usage(void)
 {
 	fprintf(stderr, "%s: usage: %s -p socketfile [options]\n"
 	                "-a peerlist \tfile containing list of hosts to ignore\n"
@@ -2709,21 +2733,18 @@ usage()
 	                "-h          \tappend identifying header\n"
 	                "-H          \tsign with explicit header lists\n"
 	                "-i ilist    \tfile containing list of internal (signing) hosts\n"
-#if _FFR_EXTERNAL_IGNORE_LIST
 	                "-I elist    \tfile containing list of external domain clients\n"
-#endif /* _FFR_EXTERNAL_IGNORE_LIST */
 #if _FFR_MULTIPLE_KEYS
 	                "-k          \tload a key set instead of a single key\n"
 #endif /* _FFR_MULTIPLE_KEYS */
 	                "-l          \tlog activity to system log\n"
 	                "-m mtalist  \tMTA daemon names for which to sign\n"
-#if _FFR_MACRO_LIST
 	                "-M macrolist\tMTA macros which enable signing\n"
-#endif /* _FFR_MACRO_LIST */
+	                "-o hdrlist  \tlist of headers to omit from signing\n"
 	                "-P pidfile  \tfile to which to write pid\n"
-#if _FFR_REQUIRE_HEADERS
+#if _FFR_REQUIRED_HEADERS
 	                "-r          \trequire basic RFC2822 header compliance\n"
-#endif /* _FFR_REQUIRE_HEADERS */
+#endif /* _FFR_REQUIRED_HEADERS */
 #if _FFR_REPORTINFO
 	                "-R          \tgenerate verification failure reports\n"
 #endif /* _FFR_REPORTINFO */
@@ -2745,9 +2766,7 @@ usage()
 */
 
 int
-main(argc, argv)
-	int argc;
-	char *argv[];
+main(int argc, char **argv)
 {
 	bool autorestart = FALSE;
 	bool gotp = FALSE;
@@ -2769,19 +2788,16 @@ main(argc, argv)
 	char *p;
 	char *pidfile = NULL;
 	char *keyfile = NULL;
-	char *config = NULL;
+	char *confstr = NULL;
 	char *peerfile = NULL;
 	char *ilist = NULL;
+	char *omitlist = NULL;
 #if POPAUTH
 	char *popdbfile = NULL;
 #endif /* POPAUTH */
-#if _FFR_EXTERNAL_IGNORE_LIST
 	char *elist = NULL;
-#endif /* _FFR_EXTERNAL_IGNORE_LIST */
 	char *canonstr = NULL;
-#if _FFR_MACRO_LIST
 	char *macrolist = NULL;
-#endif /* _FFR_MACRO_LIST */
 	unsigned char *s33krit = NULL;
 #ifndef DEBUG
 	char *end;
@@ -2793,6 +2809,8 @@ main(argc, argv)
 	dolog = FALSE;
 	subdomains = FALSE;
 	hdrlist = FALSE;
+	dompats = NULL;
+	omithdrs = NULL;
 #if POPAUTH
 	popdb = NULL;
 #endif /* POPAUTH */
@@ -2812,15 +2830,11 @@ main(argc, argv)
 	libdk = NULL;
 	domains = NULL;
 	mtas = NULL;
-#if _FFR_MACRO_LIST
 	macros = NULL;
 	values = NULL;
-#endif /* _FFR_MACRO_LIST */
 	peerlist = NULL;
 	internal = NULL;
-#if _FFR_EXTERNAL_IGNORE_LIST
 	exignore = NULL;
-#endif /* _FFR_EXTERNAL_IGNORE_LIST */
 	canon = DK_CANON_SIMPLE;
 	tmo = DEFTIMEOUT;
 	if (DK_DEBUG('t'))
@@ -2861,7 +2875,7 @@ main(argc, argv)
 		  case 'C':
 			if (optarg == NULL || *optarg == '\0')
 				return usage();
-			config = optarg;
+			confstr = optarg;
 			break;
 
 		  case 'd':
@@ -2904,13 +2918,11 @@ main(argc, argv)
 			ilist = optarg;
 			break;
 
-#if _FFR_EXTERNAL_IGNORE_LIST
 		  case 'I':
 			if (optarg == NULL || *optarg == '\0')
 				return usage();
 			elist = optarg;
 			break;
-#endif /* _FFR_EXTERNAL_IGNORE_LIST */
 
 #if _FFR_MULTIPLE_KEYS
 		  case 'k':
@@ -2930,13 +2942,17 @@ main(argc, argv)
 			mtalist = optarg;
 			break;
 
-#if _FFR_MACRO_LIST
 		  case 'M':
 			if (optarg == NULL || *optarg == '\0')
 				return usage();
 			macrolist = optarg;
 			break;
-#endif /* _FFR_MACRO_LIST */
+
+		  case 'o':
+			if (optarg == NULL || *optarg == '\0')
+				return usage();
+			omitlist = optarg;
+			break;
 
 		  case 'p':
 			if (optarg == NULL || *optarg == '\0')
@@ -3015,7 +3031,7 @@ main(argc, argv)
 	if (optind != argc)
 		return usage();
 
-	if (!dkf_parseconfig(config))
+	if (!dkf_parseconfig(confstr))
 		return EX_USAGE;
 
 	if (!gotp)
@@ -3061,17 +3077,212 @@ main(argc, argv)
 	}
 
 	if (domlist != NULL)
+	{
+		int n;
+
+		bool makepats = FALSE;
+
+		if (domlist[0] == '/' && access(domlist, F_OK) == 0)
+		{
+			int nalloc = 0;
+			FILE *f;
+			char line[BUFRSZ + 1];
+
+			n = 0;
+
+			f = fopen(domlist, "r");
+			if (f == NULL)
+			{
+				fprintf(stderr, "%s: %s: fopen(): %s\n",
+				        progname, domlist, strerror(errno));
+				return EX_UNAVAILABLE;
+			}
+
+			memset(line, '\0', sizeof line);
+			while (fgets(line, BUFRSZ, f) != NULL)
+			{
+				for (p = line; *p != '\0'; p++)
+				{
+					if (*p == '\n' || *p == '#')
+					{
+						*p = '\0';
+						break;
+					}
+				}
+
+				dkf_trimspaces(line);
+				if (strlen(line) == 0)
+					continue;
+
+				if (nalloc <= n)
+				{
+					if (nalloc == 0)
+					{
+						domains = (char **) malloc(2 * sizeof(char *));
+						nalloc = 1;
+					}
+					else
+					{
+						domains = (char **) realloc(domains,
+						                            (nalloc * 2 + 1) * sizeof(char *));
+						nalloc = nalloc * 2;
+					}
+
+					if (domains == NULL)
+					{
+						fprintf(stderr,
+						        "%s: malloc(): %s\n",
+						        progname,
+						        strerror(errno));
+						return EX_UNAVAILABLE;
+					}
+				}
+
+				domains[n] = strdup(line);
+				if (domains[n] == NULL)
+				{
+					fprintf(stderr, "%s: strdup(): %s\n",
+					        progname, strerror(errno));
+					return EX_UNAVAILABLE;
+				}
+
+				if (strchr(domains[n], '*') != NULL)
+					makepats = TRUE;
+
+				n++;
+			}
+
+			if (domains != NULL)
+				domains[n] = NULL;
+
+			fclose(f);
+		}
+		else
+		{
+			n = 1;
+
+			for (p = domlist; *p != '\0'; p++)
+			{
+				if (*p == ',')
+					n++;
+			}
+
+			domains = (char **) malloc((n + 1) * sizeof(char *));
+			if (domains == NULL)
+			{
+				fprintf(stderr, "%s: malloc(): %s\n",
+				        progname, strerror(errno));
+				return EX_UNAVAILABLE;
+			}
+
+			n = 0;
+
+			for (p = strtok(domlist, ",");
+			     p != NULL;
+			     p = strtok(NULL, ","))
+			{
+				domains[n] = p;
+
+				if (strchr(domains[n], '*') != NULL)
+					makepats = TRUE;
+
+				n++;
+			}
+
+			domains[n] = NULL;
+		}
+
+		if (makepats)
+		{
+			char *end;
+			char *q;
+			char patbuf[BUFRSZ + 1];
+
+			dompats = (regex_t **) malloc ((n + 1) * sizeof (regex_t *));
+			if (dompats == NULL)
+			{
+				fprintf(stderr, "%s: malloc(): %s\n",
+				        progname, strerror(errno));
+				return EX_UNAVAILABLE;
+			}
+			memset(dompats, '\0', (n + 1) * sizeof (regex_t *));
+
+			for (c = 0; c < n; c++)
+			{
+				memset(patbuf, '\0', sizeof patbuf);
+				end = patbuf + sizeof patbuf;
+				patbuf[0] = '^';
+
+				for (p = domains[c], q = patbuf + 1;
+				     *p != '\0' && q < end;
+				     p++)
+				{
+					switch (*p)
+					{
+					  case '*':
+						*q = '.';
+						q++;
+						*q = '*';
+						q++;
+						break;
+
+					  case '.':
+						*q = '\\';
+						q++;
+						*q = '.';
+						q++;
+						break;
+
+					  default:
+						*q = *p;
+						q++;
+						break;
+					}
+				}
+
+				*q++ = '$';
+				if (q >= end)
+				{
+					fprintf(stderr,
+					        "%s: regular expression for \"%s\" too large\n",
+					        progname, domains[c]);
+					return EX_UNAVAILABLE;
+				}
+
+				dompats[c] = (regex_t *) malloc(sizeof(regex_t));
+				if (dompats[c] == NULL)
+				{
+					fprintf(stderr, "%s: malloc(): %s\n",
+					        progname, strerror(errno));
+					return EX_UNAVAILABLE;
+				}
+
+				status = regcomp(dompats[c], patbuf,
+				                 (REG_EXTENDED|REG_ICASE));
+				if (status != 0)
+				{
+					(void) regerror(status, dompats[c],
+					                patbuf, sizeof patbuf);
+					fprintf(stderr, "%s: regcomp(): %s\n",
+					        progname, patbuf);
+					return EX_UNAVAILABLE;
+				}
+			}
+		}
+	}
+
+	if (omitlist != NULL)
 	{
 		int n = 1;
 
-		for (p = domlist; *p != '\0'; p++)
+		for (p = omitlist; *p != '\0'; p++)
 		{
 			if (*p == ',')
 				n++;
 		}
 
-		domains = (char **) malloc((n + 1) * sizeof(char *));
-		if (domains == NULL)
+		omithdrs = (char **) malloc((n + 1) * sizeof(char *));
+		if (omithdrs == NULL)
 		{
 			fprintf(stderr, "%s: malloc(): %s\n",
 			        progname, strerror(errno));
@@ -3080,18 +3291,18 @@ main(argc, argv)
 
 		n = 0;
 
-		for (p = strtok(domlist, ",");
+		for (p = strtok(omitlist, ",");
 		     p != NULL;
 		     p = strtok(NULL, ","))
-			domains[n++] = p;
+			omithdrs[n++] = p;
 
-		domains[n] = NULL;
+		omithdrs[n] = NULL;
 	}
 
-#if _FFR_MACRO_LIST
 	if (macrolist != NULL)
 	{
 		int n = 1;
+		char *macrocopy;
 
 		for (p = macrolist; *p != '\0'; p++)
 		{
@@ -3101,8 +3312,9 @@ main(argc, argv)
 
 		macros = (char **) malloc((n + 1) * sizeof(char *));
 		values = (char **) malloc((n + 1) * sizeof(char *));
+		macrocopy = strdup(macrolist);
 
-		if (macros == NULL || values == NULL)
+		if (macros == NULL || values == NULL || macrocopy == NULL)
 		{
 			fprintf(stderr, "%s: malloc(): %s\n",
 			        progname, strerror(errno));
@@ -3110,7 +3322,7 @@ main(argc, argv)
 		}
 
 		n = 0;
-		for (p = strtok(macrolist, ",");
+		for (p = strtok(macrocopy, ",");
 		     p != NULL;
 		     p = strtok(NULL, ","))
 		{
@@ -3126,7 +3338,6 @@ main(argc, argv)
 		macros[n] = NULL;
 		values[n] = NULL;
 	}
-#endif /* _FFR_MACRO_LIST */
 
 	if (mtalist != NULL)
 	{
@@ -3217,7 +3428,6 @@ main(argc, argv)
 		internal = newpeer;
 	}
 
-#if _FFR_EXTERNAL_IGNORE_LIST
 	/* external ignore list */
 	if (elist != NULL)
 	{
@@ -3239,7 +3449,6 @@ main(argc, argv)
 
 		fclose(f);
 	}
-#endif /* _FFR_EXTERNAL_IGNORE_LIST */
 
 	/* activate logging */
 	if (dolog)
@@ -3661,6 +3870,10 @@ main(argc, argv)
 		}
 	}
 
+#ifdef _FFR_LOG_SSL_ERRORS
+	ERR_load_crypto_strings();
+#endif /* _FFR_LOG_SSL_ERRORS */
+
 	/* initialize the DomainKeys package */
 	libdk = dk_init(NULL, NULL);
 	if (libdk == NULL)
@@ -3672,12 +3885,23 @@ main(argc, argv)
 		return EX_UNAVAILABLE;
 	}
 
+	if (hdrlist)
+	{
+		int opts;
+
+		opts = DK_OPTS_HDRLIST;
+
+		(void) dk_options(libdk, DK_OP_SETOPT, &opts);
+	}
+
 #if _FFR_REPORTINFO
 	if (send_reports)
 	{
 		int opts;
 
-		opts = DK_OPTS_TMPFILES;
+		(void) dk_options(libdk, DK_OP_GETOPT, &opts);
+
+		opts |= DK_OPTS_TMPFILES;
 
 		(void) dk_options(libdk, DK_OP_SETOPT, &opts);
 	}
