@@ -1,19 +1,19 @@
 /*
-**  Copyright (c) 2004, 2005 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2004-2006 Sendmail, Inc. and its suppliers.
 **	All rights reserved.
 **
-**  $Id: dk-filter.h,v 1.72 2005/04/29 00:09:26 msk Exp $
+**  $Id: dk-filter.h,v 1.86 2006/06/14 18:42:57 msk Exp $
 */
 
 #ifndef _DK_FILTER_H_
 #define _DK_FILTER_H_
 
 #ifndef lint
-static char dk_filter_h_id[] = "@(#)$Id: dk-filter.h,v 1.72 2005/04/29 00:09:26 msk Exp $";
+static char dk_filter_h_id[] = "@(#)$Id: dk-filter.h,v 1.86 2006/06/14 18:42:57 msk Exp $";
 #endif /* !lint */
 
 #define	DKF_PRODUCT	"Sendmail DomainKeys Filter"
-#define	DKF_VERSION	"0.3.0"
+#define	DKF_VERSION	"0.4.1"
 
 /* libdk */
 #include <dk.h>
@@ -28,7 +28,7 @@ static char dk_filter_h_id[] = "@(#)$Id: dk-filter.h,v 1.72 2005/04/29 00:09:26
 
 /* defaults, limits, etc. */
 #define	BUFRSZ		256
-#define CMDLINEOPTS	"a:Ab:c:C:d:DfFhHi:I:klm:M:p:P:rRs:S:T:u:U:V"
+#define CMDLINEOPTS	"a:Ab:c:C:d:DfFhHi:I:klm:M:o:p:P:rRs:S:T:u:U:V"
 #define	DEFTIMEOUT	5
 #define	HEADERMARGIN	75
 #define	HOSTUNKNOWN	"unknown-host"
