@@ -1,10 +1,10 @@
 /*
-**  Copyright (c) 2004, 2005 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2004-2006 Sendmail, Inc. and its suppliers.
 **    All rights reserved.
 */
 
 #ifndef lint
-static char ar_c_id[] = "@(#)$Id: ar.c,v 1.50 2005/04/08 17:33:23 msk Exp $";
+static char ar_c_id[] = "@(#)$Id: ar.c,v 1.58 2006/05/03 18:20:26 msk Exp $";
 #endif /* !lint */
 
 /* OS stuff */
@@ -15,6 +15,9 @@ static char ar_c_id[] = "@(#)$Id: ar.c,v 1.50 2005/04/08 17:33:23 msk Exp $";
 /* system includes */
 #include <sys/param.h>
 #include <sys/types.h>
+#ifdef SOLARIS
+# include <iso/limits_iso.h>
+#endif /* SOLARIS */
 #include <sys/socket.h>
 #include <sys/time.h>
 #include <sys/uio.h>
@@ -64,6 +67,7 @@ static char ar_c_id[] = "@(#)$Id: ar.c,v 1.50 2005/04/08 17:33:23 msk Exp $";
 
 /* ar includes */
 #include "ar.h"
+#include "fix.h"
 
 /*
 **  DATA TYPES
@@ -184,6 +188,66 @@ ar_free(AR_LIB lib, void *ptr)
 		free(ptr);
 }
 
+/*
+**  AR_TIMELEFT -- given a start time and a duration, see how much is left
+**
+**  Parameters:
+**  	start -- start time
+**  	length -- run time
+**  	remain -- how much time is left (updated)
+**
+**  Return value:
+**   	None.
+**
+**  Notes:
+**  	If "start" is NULL, "length" is taken to be the end time.
+*/
+
+static void
+ar_timeleft(struct timeval *start, struct timeval *length,
+            struct timeval *remain)
+{
+	struct timeval now;
+	struct timeval end;
+
+	assert(length != NULL);
+	assert(remain != NULL);
+
+	(void) gettimeofday(&now, NULL);
+
+	if (start == NULL)
+	{
+		memcpy(&end, length, sizeof end);
+	}
+	else
+	{
+		end.tv_sec = start->tv_sec + length->tv_sec;
+		end.tv_usec = start->tv_usec + length->tv_usec;
+		end.tv_sec += end.tv_usec / 1000000;
+		end.tv_usec = end.tv_usec % 1000000;
+	}
+
+	if (now.tv_sec > end.tv_sec ||
+	    (now.tv_sec == end.tv_sec && now.tv_usec > end.tv_usec))
+	{
+		remain->tv_sec = 0;
+		remain->tv_usec = 0;
+	}
+	else
+	{
+		remain->tv_sec = end.tv_sec - now.tv_sec;
+		if (end.tv_usec < now.tv_usec)
+		{
+			remain->tv_sec--;
+			remain->tv_usec = end.tv_usec - now.tv_usec + 1000000;
+		}
+		else
+		{
+			remain->tv_usec = end.tv_usec - now.tv_usec;
+		}
+	}
+}
+
 /*
 **  AR_ELAPSED -- determine whether or not a certain amount of time has
 **                elapsed
@@ -622,6 +686,8 @@ ar_dispatcher(void *tp)
 	fd_set rfds;
 	fd_set wfds;
 #endif /* SELECT */
+	struct timeval timeout;
+	struct timeval timeleft;
 
 	assert(tp != NULL);
 
@@ -643,13 +709,39 @@ ar_dispatcher(void *tp)
 			FD_SET(lib->ar_nsfd, &wfds);
 		FD_SET(lib->ar_nsfd, &rfds);
 		FD_SET(lib->ar_control[1], &rfds);
-#endif /* SELECT */
+
+		/* determine how long to wait */
+		timeout.tv_sec = INT_MAX;
+		timeout.tv_usec = 0;
+		for (q = lib->ar_queries; q != NULL; q = q->q_next)
+		{
+			/* check for absolute timeout */
+			if (q->q_timeout.tv_sec != 0 &&
+			    (q->q_flags & QUERY_INFINIWAIT) == 0)
+			{
+				ar_timeleft(NULL, &q->q_timeout, &timeleft);
+				if (timeleft.tv_sec < timeout.tv_sec ||
+				    (timeleft.tv_sec == timeout.tv_sec &&
+				     timeleft.tv_usec < timeout.tv_usec))
+				{
+					memcpy(&timeout, &timeleft,
+					       sizeof timeout);
+				}
+			}
+
+			/* check for re-send timeout */
+			ar_timeleft(&q->q_sent, &lib->ar_retry, &timeleft);
+			if (timeleft.tv_sec < timeout.tv_sec ||
+			    (timeleft.tv_sec == timeout.tv_sec &&
+			     timeleft.tv_usec < timeout.tv_usec))
+				memcpy(&timeout, &timeleft, sizeof timeout);
+		}
 
 		pthread_mutex_unlock(&lib->ar_lock);
 
-#if SELECT
 		/* XXX -- effect a poll if we knew there was more pending */
-		status = select(maxfd + 1, &rfds, &wfds, NULL, NULL);
+		status = select(maxfd + 1, &rfds, &wfds, NULL,
+		                lib->ar_queries != NULL ? &timeout : NULL);
 		if (status == -1)
 		{
 			if (errno == EINTR)
@@ -658,10 +750,8 @@ ar_dispatcher(void *tp)
 				assert(status >= 0);
 		}
 
-		assert(status > 0);
-#endif /* SELECT */
-
 		pthread_mutex_lock(&lib->ar_lock);
+#endif /* SELECT */
 
 		wrote = FALSE;
 
@@ -1037,7 +1127,7 @@ ar_init(ar_malloc_t user_malloc, ar_free_t user_free, void *user_closure,
 	if (new == NULL)
 		return NULL;
 
-	res_init();
+	ar_res_init();
 
 	new->ar_nscount = _res.nscount;
 	new->ar_malloc = user_malloc;
@@ -1321,7 +1411,7 @@ ar_addquery(AR_LIB lib, char *name, int class, int type, int depth,
 		q = ar_malloc(lib, sizeof(struct ar_query));
 		if (q == NULL)
 		{
-			if (*err != NULL)
+			if (err != NULL)
 				*err = errno;
 			return NULL;
 		}
