@@ -1,12 +1,12 @@
 /*
-**  Copyright (c) 2004, 2005 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2004-2006 Sendmail, Inc. and its suppliers.
 **    All rights reserved.
 **
-**  $Id: dk.c,v 1.157 2005/04/28 23:50:15 msk Exp $
+**  $Id: dk.c,v 1.168 2006/06/06 23:22:45 msk Exp $
 */
 
 #ifndef lint
-static char dk_c_id[] = "@(#)$Id: dk.c,v 1.157 2005/04/28 23:50:15 msk Exp $";
+static char dk_c_id[] = "@(#)$Id: dk.c,v 1.168 2006/06/06 23:22:45 msk Exp $";
 #endif /* !lint */
 
 /* system includes */
@@ -415,6 +415,7 @@ dk_in_use(DK *dk)
 #if USE_ARLIB
 	AR_QUERY q;
 #endif /* USE_ARLIB */
+	char *ssel;
 	unsigned char *p;
 	unsigned char *cp;
 	unsigned char *eom;
@@ -430,6 +431,10 @@ dk_in_use(DK *dk)
 
 	assert(dk != NULL);
 
+	ssel = dk_sterilize(dk->dk_domain);
+	if (ssel == NULL)
+		return -1;
+
 	/* send the NS query */
 	memset(qname, '\0', sizeof qname);
 	snprintf(qname, sizeof qname - 1, "%s.%s", DK_DNSNAME,
@@ -645,7 +650,7 @@ dk_in_use(DK *dk)
 			if (*p == ';')
 			{
 				state = 4;
-				continue;
+				/* FALLTHROUGH */
 			}
 			else if (isascii(*p) && isspace(*p))
 			{
@@ -735,6 +740,7 @@ dk_get_key(DK *dk)
 	unsigned char *p;
 	unsigned char *cp;
 	unsigned char *eom;
+	char *ssel;
 	char qname[MAXHOSTNAMELEN + 1];
 	unsigned char ansbuf[MAXPACKET];
 	char tag[MAXPARAMETER + 1];
@@ -757,10 +763,13 @@ dk_get_key(DK *dk)
 		if (dk->dk_key != NULL)
 			return DK_STAT_OK;
 
+		ssel = dk_sterilize(dk->dk_selector);
+		if (ssel == NULL)
+			return DK_STAT_SYNTAX;
+
 		memset(fn, '\0', sizeof fn);
 		snprintf(fn, sizeof fn - 1, DK_PRIVATEKEY,
-		         dk->dk_selector == NULL ? DK_DEFAULT_SELECTOR
-		                                 : dk_sterilize(dk->dk_selector));
+		         dk->dk_selector == NULL ? DK_DEFAULT_SELECTOR : ssel);
 		f = fopen(fn, "r");
 		if (f == NULL)
 			return DK_STAT_INTERNAL;
@@ -983,7 +992,7 @@ dk_get_key(DK *dk)
 			if (*p == ';')
 			{
 				state = 4;
-				continue;
+				/* FALLTHROUGH */
 			}
 			else if (isascii(*p) && isspace(*p))
 			{
@@ -1039,7 +1048,7 @@ dk_get_key(DK *dk)
 			}
 			else
 			{
-				return DK_STAT_CANTVRFY;
+				/* XXX -- unknown; ignore */
 			}
 
 			state = 0;
@@ -1337,11 +1346,7 @@ dk_process_dkheader(DK *dk, const char *hdr)
 			/* FALLTHROUGH */
 
 		  case 4:			/* in value */
-			/*
-			if (!quoting &&
-			    (*p == ';' || (isascii(*p) && isspace(*p))))
-			*/
-			if (!quoting && *p == ';')
+			if (!quoting && !escaped && *p == ';')
 			{
 				status = dk_process_value(dk, pcode, value);
 				if (status == 0)
@@ -1364,7 +1369,9 @@ dk_process_dkheader(DK *dk, const char *hdr)
 			  case DK_PARAM_QUERYMETHOD:
 			  case DK_PARAM_SELECTOR:
 			  case DK_PARAM_HDRLIST:
-				if (plen < MAXPARAMETER)
+				if (plen < MAXPARAMETER &&
+				    (quoting || escaped ||
+				     !(isascii(*p) && isspace(*p))))
 				{
 					value[plen] = *p;
 					plen++;
@@ -1500,6 +1507,8 @@ dk_new(DK_LIB *libhandle, const char *id, void *memclosure,
 	new->dk_libhandle = libhandle;
 	new->dk_tmpdir = libhandle->dkl_tmpdir;
 	new->dk_timeout = DEFTIMEOUT;
+	if ((libhandle->dkl_options & DK_OPTS_HDRLIST) != 0)
+		new->dk_flags |= DK_FLAG_HDRLIST;
 
 	*statp = DK_STAT_OK;
 	return new;
@@ -1563,7 +1572,9 @@ dk_init(void *(*caller_mallocf)(void *closure, size_t nbytes),
 	(void) res_init();
 
 	/* initialize OpenSSL algorithms */
+#if 0
 	OpenSSL_add_all_algorithms();
+#endif /* 0 */
 
 	/* copy the parameters */
 	libhandle = (DK_LIB *) malloc(sizeof(struct dk_lib));
@@ -2027,7 +2038,7 @@ dk_eoh(DK *dk)
 	**  algorithm, but this is a small data set.
 	*/
 
-	if (dk->dk_hdrlidx > 0)
+	if (dk->dk_hdrlidx > 0 && (dk->dk_flags & DK_FLAG_HDRLIST) != 0)
 	{
 		/* reset sender/from buffer to use what was actually signed */
 		memset(dk->dk_uhdrn, '\0', sizeof dk->dk_uhdrn);
@@ -2212,7 +2223,6 @@ dk_body(DK *dk, u_char *buf, size_t len)
 	u_char *p;
 	u_char *eob;
 	u_char *wrote;
-	struct dk_sha1 *sha1;
 	unsigned char lbuf[BUFRSZ];
 
 	assert(dk != NULL);
@@ -2227,7 +2237,6 @@ dk_body(DK *dk, u_char *buf, size_t len)
 	if (dk->dk_skipbody)
 		return DK_STAT_OK;
 
-	sha1 = (struct dk_sha1 *) dk->dk_signinfo;
 	eob = buf + len - 1;
 	wrote = buf;
 	wlen = 0;
@@ -2368,10 +2377,7 @@ dk_body(DK *dk, u_char *buf, size_t len)
 		}
 
 		if (lidx > 0)
-		{
 			dk_canon(dk, lbuf, lidx);
-			dk_canon(dk, CRLF, 2);
-		}
 
 		break;
 
@@ -2411,14 +2417,9 @@ dk_eom(DK *dk, DK_FLAGS *dkf)
 	{
 		int status;
 
+		/* no sender header was found below the signature */
 		if (dk->dk_skipbody)
-		{
-			(void) dk_in_use(dk);
-			if (dk->dk_signall || DK_DEBUG('s'))
-				return DK_STAT_NOSIG;
-			else
-				return DK_STAT_OK;
-		}
+			return DK_STAT_SYNTAX;
 
 		if (dk->dk_key == NULL)
 		{
@@ -2574,7 +2575,7 @@ dk_getsig(DK *dk, u_char *buf, size_t len)
 	                  sha1->sha1_rsainlen, sha1->sha1_rsaout,
 	                  &l, sha1->sha1_rsa);
 
-	if (status == 0)
+	if (status == 0 || l == 0)
 	{
 		BIO_free(key);
 		return DK_STAT_INTERNAL;
@@ -2864,11 +2865,13 @@ dk_free(DK *dk)
 
 	CLOBBER(dk->dk_sender);
 	CLOBBER(dk->dk_domain);
+	CLOBBER(dk->dk_user);
 	CLOBBER(dk->dk_selector);
 	if (dk->dk_flags & DK_FLAG_FREESIGNATURE)
 		CLOBBER(dk->dk_signature);
 	CLOBBER(dk->dk_b64);
 	CLOBBER(dk->dk_key);
+	CLOBBER(dk->dk_gran);
 
 	dk_mfree(dk->dk_libhandle, dk->dk_closure, dk);
 
